@@ -136,7 +136,7 @@ bool sequence_definition::check_and_adjust(bool set_reserved_until)
 
 void sequence_definition::read_fields(TABLE *table)
 {
-  MY_BITMAP *old_map= dbug_tmp_use_all_columns(table, &table->read_set);
+  my_bitmap_map *old_map= dbug_tmp_use_all_columns(table, table->read_set);
   reserved_until= table->field[0]->val_int();
   min_value=      table->field[1]->val_int();
   max_value=      table->field[2]->val_int();
@@ -145,7 +145,7 @@ void sequence_definition::read_fields(TABLE *table)
   cache=          table->field[5]->val_int();
   cycle=          table->field[6]->val_int();
   round=          table->field[7]->val_int();
-  dbug_tmp_restore_column_map(&table->read_set, old_map);
+  dbug_tmp_restore_column_map(table->read_set, old_map);
   used_fields= ~(uint) 0;
   print_dbug();
 }
@@ -157,7 +157,7 @@ void sequence_definition::read_fields(TABLE *table)
 
 void sequence_definition::store_fields(TABLE *table)
 {
-  MY_BITMAP *old_map= dbug_tmp_use_all_columns(table, &table->write_set);
+  my_bitmap_map *old_map= dbug_tmp_use_all_columns(table, table->write_set);
 
   /* zero possible delete markers & null bits */
   memcpy(table->record[0], table->s->default_values, table->s->null_bytes);
@@ -170,7 +170,7 @@ void sequence_definition::store_fields(TABLE *table)
   table->field[6]->store((longlong) cycle != 0, 0);
   table->field[7]->store((longlong) round, 1);
 
-  dbug_tmp_restore_column_map(&table->write_set, old_map);
+  dbug_tmp_restore_column_map(table->write_set, old_map);
   print_dbug();
 }
 
@@ -203,16 +203,6 @@ bool check_sequence_fields(LEX *lex, List<Create_field> *fields)
     reason= "Sequence tables cannot have any keys";
     goto err;
   }
-  if (lex->alter_info.check_constraint_list.elements > 0)
-  {
-    reason= "Sequence tables cannot have any constraints";
-    goto err;
-  }
-  if (lex->alter_info.flags & ALTER_ORDER)
-  {
-    reason= "ORDER BY";
-    goto err;
-  }
 
   for (field_no= 0; (field= it++); field_no++)
   {
@@ -220,8 +210,7 @@ bool check_sequence_fields(LEX *lex, List<Create_field> *fields)
     if (my_strcasecmp(system_charset_info, field_def->field_name,
                       field->field_name.str) ||
         field->flags != field_def->flags ||
-        field->type_handler() != field_def->type_handler ||
-        field->check_constraint || field->vcol_info)
+        field->type_handler() != field_def->type_handler)
     {
       reason= field->field_name.str;
       goto err;
@@ -231,8 +220,8 @@ bool check_sequence_fields(LEX *lex, List<Create_field> *fields)
 
 err:
   my_error(ER_SEQUENCE_INVALID_TABLE_STRUCTURE, MYF(0),
-           lex->select_lex.table_list.first->db.str,
-           lex->select_lex.table_list.first->table_name.str, reason);
+           lex->first_select_lex()->table_list.first->db.str,
+           lex->first_select_lex()->table_list.first->table_name.str, reason);
   DBUG_RETURN(TRUE);
 }
 
@@ -366,10 +355,8 @@ bool sequence_insert(THD *thd, LEX *lex, TABLE_LIST *org_table_list)
   seq->reserved_until= seq->start;
   error= seq->write_initial_sequence(table);
 
-  if (trans_commit_stmt(thd))
-    error= 1;
-  if (trans_commit_implicit(thd))
-    error= 1;
+  trans_commit_stmt(thd);
+  trans_commit_implicit(thd);
 
   if (!temporary_table)
   {
@@ -485,10 +472,6 @@ int SEQUENCE::read_initial_values(TABLE *table)
       if (mdl_lock_used)
         thd->mdl_context.release_lock(mdl_request.ticket);
       write_unlock(table);
-
-      if (!has_active_transaction && !thd->transaction.stmt.is_empty() &&
-          !thd->in_sub_stmt)
-        trans_commit_stmt(thd);
       DBUG_RETURN(HA_ERR_LOCK_WAIT_TIMEOUT);
     }
     DBUG_ASSERT(table->reginfo.lock_type == TL_READ);
@@ -504,12 +487,9 @@ int SEQUENCE::read_initial_values(TABLE *table)
       Doing mysql_lock_tables() may have started a read only transaction.
       If that happend, it's better that we commit it now, as a lot of
       code assumes that there is no active stmt transaction directly after
-      open_tables().
-      But we also don't want to commit the stmt transaction while in a
-      substatement, see MDEV-15977.
+      open_tables()
     */
-    if (!has_active_transaction && !thd->transaction.stmt.is_empty() &&
-        !thd->in_sub_stmt)
+    if (!has_active_transaction && !thd->transaction.stmt.is_empty())
       trans_commit_stmt(thd);
   }
   write_unlock(table);
@@ -527,11 +507,12 @@ int SEQUENCE::read_initial_values(TABLE *table)
 int SEQUENCE::read_stored_values(TABLE *table)
 {
   int error;
+  my_bitmap_map *save_read_set;
   DBUG_ENTER("SEQUENCE::read_stored_values");
 
-  MY_BITMAP *save_read_set= tmp_use_all_columns(table, &table->read_set);
+  save_read_set= tmp_use_all_columns(table, table->read_set);
   error= table->file->ha_read_first_row(table->record[0], MAX_KEY);
-  tmp_restore_column_map(&table->read_set, save_read_set);
+  tmp_restore_column_map(table->read_set, save_read_set);
 
   if (unlikely(error))
   {
