@@ -32,7 +32,7 @@ Created 03/11/2014 Shaohua Wang
 #include "trx0trx.h"
 
 /** Innodb B-tree index fill factor for bulk load. */
-long	innobase_fill_factor;
+uint	innobase_fill_factor;
 /** whether to reduce redo logging during ALTER TABLE */
 my_bool	innodb_log_optimize_ddl;
 
@@ -56,7 +56,7 @@ PageBulk::init()
 		m_mtr.set_log_mode(MTR_LOG_NO_REDO);
 		m_mtr.set_flush_observer(m_flush_observer);
 	} else {
-		m_mtr.set_named_space(m_index->space);
+		m_index->set_modified(m_mtr);
 	}
 
 	if (m_page_no == FIL_NULL) {
@@ -67,11 +67,12 @@ PageBulk::init()
 		the allocation order, and we will always generate redo log
 		for page allocation, even when creating a new tablespace. */
 		alloc_mtr.start();
-		alloc_mtr.set_named_space(m_index->space);
+		m_index->set_modified(alloc_mtr);
 
 		ulint	n_reserved;
 		bool	success;
-		success = fsp_reserve_free_extents(&n_reserved, m_index->space,
+		success = fsp_reserve_free_extents(&n_reserved,
+						   m_index->table->space,
 						   1, FSP_NORMAL, &alloc_mtr);
 		if (!success) {
 			alloc_mtr.commit();
@@ -83,10 +84,7 @@ PageBulk::init()
 		new_block = btr_page_alloc(m_index, 0, FSP_UP, m_level,
 					   &alloc_mtr, &m_mtr);
 
-		if (n_reserved > 0) {
-			fil_space_release_free_extents(m_index->space,
-						       n_reserved);
-		}
+		m_index->table->space->release_free_extents(n_reserved);
 
 		alloc_mtr.commit();
 
@@ -96,7 +94,7 @@ PageBulk::init()
 
 		if (new_page_zip) {
 			page_create_zip(new_block, m_index, m_level, 0,
-					NULL, &m_mtr);
+					&m_mtr);
 			memset(FIL_PAGE_PREV + new_page, 0xff, 8);
 			page_zip_write_header(new_page_zip,
 					      FIL_PAGE_PREV + new_page,
@@ -121,11 +119,10 @@ PageBulk::init()
 				       m_index->id, &m_mtr);
 		}
 	} else {
-		page_id_t	page_id(dict_index_get_space(m_index), m_page_no);
-		page_size_t	page_size(dict_table_page_size(m_index->table));
-
-		new_block = btr_block_get(page_id, page_size,
-					  RW_X_LATCH, m_index, &m_mtr);
+		new_block = btr_block_get(
+			page_id_t(m_index->table->space_id, m_page_no),
+			m_index->table->space->zip_size(),
+			RW_X_LATCH, m_index, &m_mtr);
 
 		new_page = buf_block_get_frame(new_block);
 		new_page_zip = buf_block_get_page_zip(new_block);
@@ -156,11 +153,11 @@ PageBulk::init()
 		m_reserved_space = dict_index_get_space_reserve();
 	} else {
 		m_reserved_space =
-			UNIV_PAGE_SIZE * (100 - innobase_fill_factor) / 100;
+			srv_page_size * (100 - innobase_fill_factor) / 100;
 	}
 
 	m_padding_space =
-		UNIV_PAGE_SIZE - dict_index_zip_pad_optimal_page_size(m_index);
+		srv_page_size - dict_index_zip_pad_optimal_page_size(m_index);
 	m_heap_top = page_header_get_ptr(new_page, PAGE_HEAP_TOP);
 	m_rec_no = page_header_get_field(new_page, PAGE_N_RECS);
 
@@ -185,13 +182,14 @@ PageBulk::insert(
 	ut_ad(m_heap != NULL);
 
 	rec_size = rec_offs_size(offsets);
+	ut_d(const bool is_leaf = page_rec_is_leaf(m_cur_rec));
 
 #ifdef UNIV_DEBUG
 	/* Check whether records are in order. */
 	if (!page_rec_is_infimum(m_cur_rec)) {
 		rec_t*	old_rec = m_cur_rec;
 		ulint*	old_offsets = rec_get_offsets(
-			old_rec, m_index, NULL,	page_rec_is_leaf(old_rec),
+			old_rec, m_index, NULL,	is_leaf,
 			ULINT_UNDEFINED, &m_heap);
 
 		ut_ad(cmp_rec_rec(rec, old_rec, offsets, old_offsets, m_index)
@@ -203,7 +201,7 @@ PageBulk::insert(
 
 	/* 1. Copy the record to page. */
 	rec_t*	insert_rec = rec_copy(m_heap_top, rec, offsets);
-	rec_offs_make_valid(insert_rec, m_index, offsets);
+	rec_offs_make_valid(insert_rec, m_index, is_leaf, offsets);
 
 	/* 2. Insert the record in the linked list. */
 	rec_t*	next_rec = page_rec_get_next(m_cur_rec);
@@ -229,7 +227,7 @@ PageBulk::insert(
 		- page_dir_calc_reserved_space(m_rec_no);
 
 	ut_ad(m_free_space >= rec_size + slot_size);
-	ut_ad(m_heap_top + rec_size < m_page + UNIV_PAGE_SIZE);
+	ut_ad(m_heap_top + rec_size < m_page + srv_page_size);
 
 	m_free_space -= rec_size + slot_size;
 	m_heap_top += rec_size;
@@ -306,6 +304,7 @@ PageBulk::finish()
 	page_dir_slot_set_n_owned(slot, NULL, count + 1);
 
 	ut_ad(!dict_index_is_spatial(m_index));
+	ut_ad(!page_get_instant(m_page));
 
 	if (!m_flush_observer && !m_page_zip) {
 		mlog_write_ulint(PAGE_HEADER + PAGE_N_DIR_SLOTS + m_page,
@@ -322,7 +321,7 @@ PageBulk::finish()
 		mlog_write_ulint(PAGE_HEADER + PAGE_LAST_INSERT + m_page,
 				 ulint(m_cur_rec - m_page),
 				 MLOG_2BYTES, &m_mtr);
-		mlog_write_ulint(PAGE_HEADER + PAGE_DIRECTION + m_page,
+		mlog_write_ulint(PAGE_HEADER + PAGE_DIRECTION_B - 1 + m_page,
 				 PAGE_RIGHT, MLOG_2BYTES, &m_mtr);
 		mlog_write_ulint(PAGE_HEADER + PAGE_N_DIRECTION + m_page, 0,
 				 MLOG_2BYTES, &m_mtr);
@@ -339,7 +338,7 @@ PageBulk::finish()
 		mach_write_to_2(PAGE_HEADER + PAGE_N_RECS + m_page, m_rec_no);
 		mach_write_to_2(PAGE_HEADER + PAGE_LAST_INSERT + m_page,
 				ulint(m_cur_rec - m_page));
-		mach_write_to_2(PAGE_HEADER + PAGE_DIRECTION + m_page,
+		mach_write_to_2(PAGE_HEADER + PAGE_DIRECTION_B - 1 + m_page,
 				PAGE_RIGHT);
 		mach_write_to_2(PAGE_HEADER + PAGE_N_DIRECTION + m_page, 0);
 	}
@@ -375,7 +374,7 @@ PageBulk::compress()
 	ut_ad(m_page_zip != NULL);
 
 	return(page_zip_compress(m_page_zip, m_page, m_index,
-				 page_zip_level, NULL, &m_mtr));
+				 page_zip_level, &m_mtr));
 }
 
 /** Get node pointer
@@ -509,15 +508,14 @@ PageBulk::copyOut(
 				  page_rec_is_leaf(split_rec),
 				  ULINT_UNDEFINED, &m_heap);
 
-	m_free_space += rec_get_end(last_rec, offsets)
-		- m_heap_top
+	m_free_space += ulint(rec_get_end(last_rec, offsets) - m_heap_top)
 		+ page_dir_calc_reserved_space(m_rec_no)
 		- page_dir_calc_reserved_space(n);
-	ut_ad(m_free_space > 0);
+	ut_ad(lint(m_free_space) > 0);
 	m_rec_no = n;
 
 #ifdef UNIV_DEBUG
-	m_total_data -= rec_get_end(last_rec, offsets) - m_heap_top;
+	m_total_data -= ulint(rec_get_end(last_rec, offsets) - m_heap_top);
 #endif /* UNIV_DEBUG */
 }
 
@@ -591,8 +589,9 @@ PageBulk::needExt(
 	const dtuple_t*		tuple,
 	ulint			rec_size)
 {
-	return(page_zip_rec_needs_ext(rec_size, m_is_comp,
-		dtuple_get_n_fields(tuple), m_block->page.size));
+	return page_zip_rec_needs_ext(rec_size, m_is_comp,
+				      dtuple_get_n_fields(tuple),
+				      m_block->zip_size());
 }
 
 /** Store external record
@@ -658,15 +657,15 @@ PageBulk::latch()
 		m_mtr.set_log_mode(MTR_LOG_NO_REDO);
 		m_mtr.set_flush_observer(m_flush_observer);
 	} else {
-		m_mtr.set_named_space(m_index->space);
+		m_index->set_modified(m_mtr);
 	}
 
 	/* In case the block is S-latched by page_cleaner. */
 	if (!buf_page_optimistic_get(RW_X_LATCH, m_block, m_modify_clock,
 				     __FILE__, __LINE__, &m_mtr)) {
-		m_block = buf_page_get_gen(page_id_t(m_index->space,
+		m_block = buf_page_get_gen(page_id_t(m_index->table->space_id,
 						     m_page_no),
-					   univ_page_size, RW_X_LATCH,
+					   0, RW_X_LATCH,
 					   m_block, BUF_GET_IF_IN_POOL,
 					   __FILE__, __LINE__, &m_mtr, &m_err);
 
@@ -782,7 +781,7 @@ BtrBulk::pageCommit(
 /** Log free check */
 inline void BtrBulk::logFreeCheck()
 {
-	if (log_sys->check_flush_or_checkpoint) {
+	if (log_sys.check_flush_or_checkpoint) {
 		release();
 
 		log_free_check();
@@ -976,7 +975,7 @@ BtrBulk::finish(dberr_t	err)
 {
 	ulint		last_page_no = FIL_NULL;
 
-	ut_ad(!dict_table_is_temporary(m_index->table));
+	ut_ad(!m_index->table->is_temporary());
 
 	if (m_page_bulks.size() == 0) {
 		/* The table is empty. The root page of the index tree
@@ -1008,30 +1007,27 @@ BtrBulk::finish(dberr_t	err)
 		rec_t*		first_rec;
 		mtr_t		mtr;
 		buf_block_t*	last_block;
-		page_t*		last_page;
-		page_id_t	page_id(dict_index_get_space(m_index),
-					last_page_no);
-		page_size_t	page_size(dict_table_page_size(m_index->table));
-		ulint		root_page_no = dict_index_get_page(m_index);
 		PageBulk	root_page_bulk(m_index, m_trx->id,
-					       root_page_no, m_root_level,
+					       m_index->page, m_root_level,
 					       m_flush_observer);
 
-		mtr_start(&mtr);
-		mtr.set_named_space(dict_index_get_space(m_index));
-		mtr_x_lock(dict_index_get_lock(m_index), &mtr);
+		mtr.start();
+		m_index->set_modified(mtr);
+		mtr_x_lock(&m_index->lock, &mtr);
 
 		ut_ad(last_page_no != FIL_NULL);
-		last_block = btr_block_get(page_id, page_size,
-					   RW_X_LATCH, m_index, &mtr);
-		last_page = buf_block_get_frame(last_block);
-		first_rec = page_rec_get_next(page_get_infimum_rec(last_page));
+		last_block = btr_block_get(
+			page_id_t(m_index->table->space_id, last_page_no),
+			m_index->table->space->zip_size(),
+			RW_X_LATCH, m_index, &mtr);
+		first_rec = page_rec_get_next(
+			page_get_infimum_rec(last_block->frame));
 		ut_ad(page_rec_is_user_rec(first_rec));
 
 		/* Copy last page to root page. */
 		err = root_page_bulk.init();
 		if (err != DB_SUCCESS) {
-			mtr_commit(&mtr);
+			mtr.commit();
 			return(err);
 		}
 		root_page_bulk.copyIn(first_rec);
@@ -1042,7 +1038,7 @@ BtrBulk::finish(dberr_t	err)
 		/* Do not flush the last page. */
 		last_block->page.flush_observer = NULL;
 
-		mtr_commit(&mtr);
+		mtr.commit();
 
 		err = pageCommit(&root_page_bulk, NULL, false);
 		ut_ad(err == DB_SUCCESS);
@@ -1051,6 +1047,6 @@ BtrBulk::finish(dberr_t	err)
 	ut_ad(!sync_check_iterate(dict_sync_check()));
 
 	ut_ad(err != DB_SUCCESS
-	      || btr_validate_index(m_index, NULL, false) == DB_SUCCESS);
+	      || btr_validate_index(m_index, NULL) == DB_SUCCESS);
 	return(err);
 }
